@@ -370,11 +370,39 @@ SerdStatus
 serd_uri_parse(const uint8_t* utf8, SerdURI* out);
 
 /**
-   Set `out` to `uri` resolved against `base`.
+   Set target `t` to reference `r` resolved against `base`.
+
+   @see http://tools.ietf.org/html/rfc3986#section-5.2.2
 */
 SERD_API
 void
-serd_uri_resolve(const SerdURI* uri, const SerdURI* base, SerdURI* out);
+serd_uri_resolve(const SerdURI* r, const SerdURI* base, SerdURI* t);
+
+/**
+   Function to detect I/O stream errors.
+
+   Identical semantics to `ferror`.
+
+   @return Non-zero if `stream` has encountered an error.
+*/
+typedef int (*SerdStreamErrorFunc)(void* stream);
+
+/**
+   Source function for raw string input.
+
+   Identical semantics to `fread`, but may set errno for more informative error
+   reporting than supported by SerdStreamErrorFunc.
+
+   @param buf Output buffer.
+   @param size Size of a single element of data in bytes (always 1).
+   @param nmemb Number of elements to read.
+   @param stream Stream to read from (FILE* for fread).
+   @return Number of elements (bytes) read.
+*/
+typedef size_t (*SerdSource)(void*  buf,
+                             size_t size,
+                             size_t nmemb,
+                             void*  stream);
 
 /**
    Sink function for raw string output.
@@ -420,6 +448,16 @@ SERD_API
 SerdNode
 serd_node_from_string(SerdType type, const uint8_t* str);
 
+/**
+   Make a (shallow) node from a prefix of `str`.
+
+   This measures, but does not copy, `str`.  No memory is allocated.
+   Note that the returned node may not be null terminated.
+*/
+SERD_API
+SerdNode
+serd_node_from_substring(SerdType type, const uint8_t* str, size_t len);
+
 /**
    Make a deep copy of `node`.
 
@@ -740,7 +778,7 @@ SERD_API
 void
 serd_reader_set_error_sink(SerdReader*   reader,
                            SerdErrorSink error_sink,
-                           void*         handle);
+                           void*         error_handle);
 
 /**
    Return the `handle` passed to serd_reader_new().
@@ -793,11 +831,26 @@ serd_reader_read_file(SerdReader*    reader,
 */
 SERD_API
 SerdStatus
-serd_reader_start_stream(SerdReader*    me,
+serd_reader_start_stream(SerdReader*    reader,
                          FILE*          file,
                          const uint8_t* name,
                          bool           bulk);
 
+/**
+   Start an incremental read from a user-specified source.
+
+   The `read_func` is guaranteed to only be called for `page_size` elements
+   with size 1 (i.e. `page_size` bytes).
+*/
+SERD_API
+SerdStatus
+serd_reader_start_source_stream(SerdReader*         reader,
+                                SerdSource          read_func,
+                                SerdStreamErrorFunc error_func,
+                                void*               stream,
+                                const uint8_t*      name,
+                                size_t              page_size);
+
 /**
    Read a single "chunk" of data during an incremental read.
 
@@ -808,14 +861,14 @@ serd_reader_start_stream(SerdReader*    me,
 */
 SERD_API
 SerdStatus
-serd_reader_read_chunk(SerdReader* me);
+serd_reader_read_chunk(SerdReader* reader);
 
 /**
    Finish an incremental read from a file handle.
 */
 SERD_API
 SerdStatus
-serd_reader_end_stream(SerdReader* me);
+serd_reader_end_stream(SerdReader* reader);
 
 /**
    Read `file`.
@@ -826,12 +879,24 @@ serd_reader_read_file_handle(SerdReader*    reader,
                              FILE*          file,
                              const uint8_t* name);
 
+/**
+   Read a user-specified byte source.
+*/
+SERD_API
+SerdStatus
+serd_reader_read_source(SerdReader*         reader,
+                        SerdSource          source,
+                        SerdStreamErrorFunc error,
+                        void*               stream,
+                        const uint8_t*      name,
+                        size_t              page_size);
+
 /**
    Read `utf8`.
 */
 SERD_API
 SerdStatus
-serd_reader_read_string(SerdReader* me, const uint8_t* utf8);
+serd_reader_read_string(SerdReader* reader, const uint8_t* utf8);
 
 /**
    Free `reader`.
@@ -855,7 +920,7 @@ serd_writer_new(SerdSyntax     syntax,
                 SerdStyle      style,
                 SerdEnv*       env,
                 const SerdURI* base_uri,
-                SerdSink       sink,
+                SerdSink       ssink,
                 void*          stream);
 
 /**
@@ -914,7 +979,7 @@ SERD_API
 void
 serd_writer_set_error_sink(SerdWriter*   writer,
                            SerdErrorSink error_sink,
-                           void*         handle);
+                           void*         error_handle);
 
 /**
    Set a prefix to be removed from matching blank node identifiers.
@@ -973,8 +1038,8 @@ serd_writer_write_statement(SerdWriter*        writer,
                             const SerdNode*    subject,
                             const SerdNode*    predicate,
                             const SerdNode*    object,
-                            const SerdNode*    object_datatype,
-                            const SerdNode*    object_lang);
+                            const SerdNode*    datatype,
+                            const SerdNode*    lang);
 
 /**
    Mark the end of an anonymous node's description.
