@@ -1,5 +1,5 @@
 /*
-  Copyright 2011-2015 David Robillard <http://drobilla.net>
+  Copyright 2011-2016 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -23,22 +23,42 @@
 #include <float.h>
 
 #ifdef _WIN32
-#    define isnan(x) _isnan(x)
-#    define isinf(x) (!_finite(x))
+#    ifndef isnan
+#        define isnan(x) _isnan(x)
+#    endif
+#    ifndef isinf
+#        define isinf(x) (!_finite(x))
+#    endif
 #endif
 
 SERD_API
 SerdNode
-serd_node_from_string(SerdType type, const uint8_t* buf)
+serd_node_from_string(SerdType type, const uint8_t* str)
 {
-	if (!buf) {
+	if (!str) {
 		return SERD_NODE_NULL;
 	}
 
 	uint32_t     flags       = 0;
 	size_t       buf_n_bytes = 0;
-	const size_t buf_n_chars = serd_strlen(buf, &buf_n_bytes, &flags);
-	SerdNode ret = { buf, buf_n_bytes, buf_n_chars, flags, type };
+	const size_t buf_n_chars = serd_strlen(str, &buf_n_bytes, &flags);
+	SerdNode ret = { str, buf_n_bytes, buf_n_chars, flags, type };
+	return ret;
+}
+
+SERD_API
+SerdNode
+serd_node_from_substring(SerdType type, const uint8_t* str, const size_t len)
+{
+	if (!str) {
+		return SERD_NODE_NULL;
+	}
+
+	uint32_t     flags       = 0;
+	size_t       buf_n_bytes = 0;
+	const size_t buf_n_chars = serd_substrlen(str, len, &buf_n_bytes, &flags);
+	assert(buf_n_bytes <= len);
+	SerdNode ret = { str, buf_n_bytes, buf_n_chars, flags, type };
 	return ret;
 }
 
@@ -114,12 +134,10 @@ serd_node_new_uri_from_string(const uint8_t* str,
                               SerdURI*       out)
 {
 	if (!str || str[0] == '\0') {
-		if (base) {
-			return serd_node_new_uri(base, NULL, out);  // Empty URI => Base
-		} else {
-			return SERD_NODE_NULL;  // Nonsense
-		}
+		// Empty URI => Base URI, or nothing if no base is given
+		return base ? serd_node_new_uri(base, NULL, out) : SERD_NODE_NULL;
 	}
+
 	SerdURI uri;
 	serd_uri_parse(str, &uri);
 	return serd_node_new_uri(&uri, base, out);  // Resolve/Serialise
@@ -185,7 +203,7 @@ serd_node_new_file_uri(const uint8_t* path,
 		serd_uri_parse(chunk.buf, out);
 	}
 
-	return serd_node_from_string(SERD_URI, chunk.buf);
+	return serd_node_from_substring(SERD_URI, chunk.buf, chunk.len);
 }
 
 SERD_API
@@ -197,17 +215,41 @@ serd_node_new_uri(const SerdURI* uri, const SerdURI* base, SerdURI* out)
 		serd_uri_resolve(uri, base, &abs_uri);
 	}
 
-	const size_t len = serd_uri_string_length(&abs_uri);
-	uint8_t*     buf = (uint8_t*)malloc(len + 1);
-
-	SerdNode node = { buf, len, len, 0, SERD_URI };  // FIXME: UTF-8
-
+	const size_t len        = serd_uri_string_length(&abs_uri);
+	uint8_t*     buf        = (uint8_t*)malloc(len + 1);
+	SerdNode     node       = { buf, 0, 0, 0, SERD_URI };
 	uint8_t*     ptr        = buf;
 	const size_t actual_len = serd_uri_serialise(&abs_uri, string_sink, &ptr);
 
 	buf[actual_len] = '\0';
 	node.n_bytes    = actual_len;
-	node.n_chars    = actual_len;
+	node.n_chars    = serd_strlen(buf, NULL, NULL);
+
+	if (out) {
+		serd_uri_parse(buf, out);  // TODO: cleverly avoid double parse
+	}
+
+	return node;
+}
+
+SERD_API
+SerdNode
+serd_node_new_relative_uri(const SerdURI* uri,
+                           const SerdURI* base,
+                           const SerdURI* root,
+                           SerdURI*       out)
+{
+	const size_t uri_len  = serd_uri_string_length(uri);
+	const size_t base_len = serd_uri_string_length(base);
+	uint8_t*     buf        = (uint8_t*)malloc(uri_len + base_len + 1);
+	SerdNode     node       = { buf, 0, 0, 0, SERD_URI };
+	uint8_t*     ptr        = buf;
+	const size_t actual_len = serd_uri_serialise_relative(
+		uri, base, root, string_sink, &ptr);
+
+	buf[actual_len] = '\0';
+	node.n_bytes    = actual_len;
+	node.n_chars    = serd_strlen(buf, NULL, NULL);
 
 	if (out) {
 		serd_uri_parse(buf, out);  // TODO: cleverly avoid double parse
