@@ -74,23 +74,17 @@ struct SerdReaderImpl {
 	Ref               rdf_rest;
 	Ref               rdf_nil;
 	SerdNode          default_graph;
-	FILE*             fd;
+	SerdByteSource    source;
 	SerdStack         stack;
 	SerdSyntax        syntax;
+	unsigned          next_id;
 	Cursor            cur;
+	SerdStatus        status;
 	uint8_t*          buf;
 	uint8_t*          bprefix;
 	size_t            bprefix_len;
-	unsigned          next_id;
-	const uint8_t*    read_buf;
-	uint8_t*          file_buf;
-	int32_t           read_head;  ///< Offset into read_buf
-	uint8_t           read_byte;  ///< 1-byte 'buffer' used when not paging
-	bool              from_file;  ///< True iff reading from `fd`
-	bool              paging;     ///< True iff reading a page at a time
 	bool              strict;     ///< True iff strict parsing
 	bool              eof;
-	bool              error;
 	bool              seen_genid;
 #ifdef SERD_STACK_CHECK
 	Ref*              allocs;     ///< Stack of push offsets
@@ -123,32 +117,23 @@ r_err(SerdReader* reader, SerdStatus st, const char* fmt, ...)
 	return 0;
 }
 
-static inline SerdStatus
-page(SerdReader* reader)
+/** fread-like wrapper for getc (which is faster). */
+static size_t
+serd_file_read_byte(void* buf, size_t size, size_t nmemb, void* stream)
 {
-	reader->read_head = 0;
-	size_t n_read = fread(reader->file_buf, 1, SERD_PAGE_SIZE, reader->fd);
-	if (n_read == 0) {
-		reader->file_buf[0] = '\0';
-		reader->eof         = true;
-		if (ferror(reader->fd)) {
-			reader->error = true;
-			r_err(reader, SERD_ERR_UNKNOWN,
-			      "read error: %s\n", strerror(errno));
-			return SERD_ERR_UNKNOWN;
-		} else {
-			return SERD_FAILURE;
-		}
-	} else if (n_read < SERD_PAGE_SIZE) {
-		reader->file_buf[n_read] = '\0';
+	const int c = getc((FILE*)stream);
+	if (c == EOF) {
+		*((uint8_t*)buf) = 0;
+		return 0;
 	}
-	return SERD_SUCCESS;
+	*((uint8_t*)buf) = (uint8_t)c;
+	return 1;
 }
 
 static inline uint8_t
 peek_byte(SerdReader* reader)
 {
-	return reader->read_buf[reader->read_head];
+	return serd_byte_source_peek(&reader->source);
 }
 
 static inline uint8_t
@@ -161,15 +146,7 @@ eat_byte_safe(SerdReader* reader, const uint8_t byte)
 	default:   ++reader->cur.col;
 	}
 
-	if (reader->from_file && !reader->paging) {
-		const int c = fgetc(reader->fd);
-		reader->read_byte = (c == EOF) ? 0 : (uint8_t)c;
-		if (c == EOF) {
-			reader->eof = true;
-		}
-	} else if (++reader->read_head == SERD_PAGE_SIZE && reader->paging) {
-		page(reader);
-	}
+	reader->status = serd_byte_source_advance(&reader->source);
 	return byte;
 }
 
@@ -212,7 +189,7 @@ push_node_padded(SerdReader* reader, size_t maxlen,
 
 #ifdef SERD_STACK_CHECK
 	reader->allocs = realloc(
-		reader->allocs, sizeof(uint8_t*) * (++reader->n_allocs));
+		reader->allocs, sizeof(reader->allocs) * (++reader->n_allocs));
 	reader->allocs[reader->n_allocs - 1] = ((uint8_t*)mem - reader->stack.buf);
 #endif
 	return (uint8_t*)node - reader->stack.buf;
@@ -301,10 +278,9 @@ read_HEX(SerdReader* reader)
 	const uint8_t c = peek_byte(reader);
 	if (is_digit(c) || in_range(c, 'A', 'F') || in_range(c, 'a', 'f')) {
 		return eat_byte_safe(reader, c);
-	} else {
-		return r_err(reader, SERD_ERR_BAD_SYNTAX,
-		             "invalid hexadecimal digit `%c'\n", c);
 	}
+	return r_err(reader, SERD_ERR_BAD_SYNTAX,
+	             "invalid hexadecimal digit `%c'\n", c);
 }
 
 // Read UCHAR escape, initial \ is already eaten by caller
@@ -427,20 +403,14 @@ bad_char(SerdReader* reader, Ref dest, const char* fmt, uint8_t c)
 		b = peek_byte(reader);
 	}
 
-	return SERD_SUCCESS;
+	return SERD_FAILURE;
 }
 
 static SerdStatus
 read_utf8_character(SerdReader* reader, Ref dest, uint8_t c)
 {
-	unsigned size = 1;
-	if ((c & 0xE0) == 0xC0) {  // Starts with `110'
-		size = 2;
-	} else if ((c & 0xF0) == 0xE0) {  // Starts with `1110'
-		size = 3;
-	} else if ((c & 0xF8) == 0xF0) {  // Starts with `11110'
-		size = 4;
-	} else {
+	const uint32_t size = utf8_num_bytes(c);
+	if (size <= 1 || size > 4) {
 		return bad_char(reader, dest, "invalid UTF-8 start 0x%X\n", c);
 	}
 
@@ -479,9 +449,8 @@ read_character(SerdReader* reader, Ref dest, SerdNodeFlags* flags, uint8_t c)
 		}
 		push_byte(reader, dest, c);
 		return SERD_SUCCESS;
-	} else {
-		return read_utf8_character(reader, dest, c);
 	}
+	return read_utf8_character(reader, dest, c);
 }
 
 // [10] comment ::= '#' ( [^#xA #xD] )*
@@ -542,7 +511,7 @@ static Ref
 read_STRING_LITERAL_LONG(SerdReader* reader, SerdNodeFlags* flags, uint8_t q)
 {
 	Ref ref = push_node(reader, SERD_LITERAL, "", 0);
-	while (true) {
+	while (!reader->status) {
 		const uint8_t c = peek_byte(reader);
 		uint32_t      code;
 		switch (c) {
@@ -563,11 +532,10 @@ read_STRING_LITERAL_LONG(SerdReader* reader, SerdNodeFlags* flags, uint8_t q)
 				if (q2 == q && q3 == q) {  // End of string
 					eat_byte_safe(reader, q3);
 					return ref;
-				} else {
-					*flags |= SERD_HAS_QUOTE;
-					push_byte(reader, ref, c);
-					read_character(reader, ref, flags, q2);
 				}
+				*flags |= SERD_HAS_QUOTE;
+				push_byte(reader, ref, c);
+				read_character(reader, ref, flags, q2);
 			} else {
 				read_character(reader, ref, flags, eat_byte_safe(reader, c));
 			}
@@ -582,9 +550,9 @@ static Ref
 read_STRING_LITERAL(SerdReader* reader, SerdNodeFlags* flags, uint8_t q)
 {
 	Ref ref = push_node(reader, SERD_LITERAL, "", 0);
-	while (true) {
-		const uint8_t c = peek_byte(reader);
-		uint32_t      code;
+	while (!reader->status) {
+		const uint8_t c    = peek_byte(reader);
+		uint32_t      code = 0;
 		switch (c) {
 		case '\n': case '\r':
 			r_err(reader, SERD_ERR_BAD_SYNTAX, "line end in short string\n");
@@ -695,11 +663,10 @@ read_PLX(SerdReader* reader, Ref dest)
 		if (is_alpha(c = peek_byte(reader))) {
 			// Escapes like \u \n etc. are not supported
 			return SERD_ERR_BAD_SYNTAX;
-		} else {
-			// Allow escaping of pretty much any other character
-			push_byte(reader, dest, eat_byte_safe(reader, c));
-			return SERD_SUCCESS;
 		}
+		// Allow escaping of pretty much any other character
+		push_byte(reader, dest, eat_byte_safe(reader, c));
+		return SERD_SUCCESS;
 	default:
 		return SERD_FAILURE;
 	}
@@ -796,45 +763,52 @@ read_LANGTAG(SerdReader* reader)
 	return ref;
 }
 
-typedef enum { PREFIX, GOOD, BAD} SchemeState;
-
-static inline bool
-check_scheme(SerdReader* reader, uint8_t c, SchemeState* state)
+static bool
+read_IRIREF_scheme(SerdReader* reader, Ref dest)
 {
-	if (!supports_relative_iris(reader) && *state == PREFIX) {
-		if (c == ':') {
-			*state = GOOD;
-		} else if (!isalpha(c)) {
-			*state = BAD;
+	uint8_t c = peek_byte(reader);
+	if (!isalpha(c)) {
+		return r_err(reader, SERD_ERR_BAD_SYNTAX,
+		             "bad IRI scheme start `%c'\n", c);
+	}
+
+	while ((c = peek_byte(reader))) {
+		if (c == '>') {
+			return r_err(reader, SERD_ERR_BAD_SYNTAX, "missing IRI scheme\n");
+		} else if (!is_uri_scheme_char(c)) {
 			return r_err(reader, SERD_ERR_BAD_SYNTAX,
-			             "syntax does not support relative IRIs\n");
+			             "bad IRI scheme char `%X'\n", c);
+		}
+
+		push_byte(reader, dest, eat_byte_safe(reader, c));
+		if (c == ':') {
+			return true;  // End of scheme
 		}
 	}
-	return true;
+
+	return false;
 }
 
 static Ref
 read_IRIREF(SerdReader* reader)
 {
 	TRY_RET(eat_byte_check(reader, '<'));
-	Ref         ref    = push_node(reader, SERD_URI, "", 0);
-	SchemeState scheme = PREFIX;
-	uint32_t    code;
-	while (true) {
-		const uint8_t c = peek_byte(reader);
-		if (!check_scheme(reader, c, &scheme)) {
-			return pop_node(reader, ref);
-		}
+	Ref ref = push_node(reader, SERD_URI, "", 0);
+	if (!supports_relative_iris(reader) && !read_IRIREF_scheme(reader, ref)) {
+		return pop_node(reader, ref);
+	}
+
+	uint32_t code = 0;
+	while (!reader->status) {
+		const uint8_t c = eat_byte_safe(reader, peek_byte(reader));
 		switch (c) {
 		case '"': case '<': case '^': case '`': case '{': case '|': case '}':
 			r_err(reader, SERD_ERR_BAD_SYNTAX,
 			      "invalid IRI character `%c'\n", c);
 			return pop_node(reader, ref);
 		case '>':
-			eat_byte_safe(reader, c);
 			return ref;
 		case '\\':
-			eat_byte_safe(reader, c);
 			if (!read_UCHAR(reader, ref, &code)) {
 				r_err(reader, SERD_ERR_BAD_SYNTAX, "invalid IRI escape\n");
 				return pop_node(reader, ref);
@@ -858,12 +832,19 @@ read_IRIREF(SerdReader* reader)
 				if (reader->strict) {
 					return pop_node(reader, ref);
 				}
-				push_byte(reader, ref, eat_byte_safe(reader, c));
-			} else {
-				push_byte(reader, ref, eat_byte_safe(reader, c));
+				reader->status = SERD_FAILURE;
+				push_byte(reader, ref, c);
+			} else if (!(c & 0x80)) {
+				push_byte(reader, ref, c);
+			} else if (read_utf8_character(reader, ref, c)) {
+				if (reader->strict) {
+					return pop_node(reader, ref);
+				}
+				reader->status = SERD_FAILURE;
 			}
 		}
 	}
+	return pop_node(reader, ref);
 }
 
 static bool
@@ -1015,27 +996,26 @@ read_verb(SerdReader* reader, Ref* dest)
 {
 	if (peek_byte(reader) == '<') {
 		return (*dest = read_IRIREF(reader));
-	} else {
-		/* Either a qname, or "a".  Read the prefix first, and if it is in fact
-		   "a", produce that instead.
-		*/
-		*dest = push_node(reader, SERD_CURIE, "", 0);
-		SerdNode*        node    = deref(reader, *dest);
-		const SerdStatus st      = read_PN_PREFIX(reader, *dest);
-		bool             ate_dot = false;
-		if (!st && node->n_bytes == 1 && node->buf[0] == 'a' &&
-		    is_token_end(peek_byte(reader))) {
-			pop_node(reader, *dest);
-			return (*dest = push_node(reader, SERD_URI, NS_RDF "type", 47));
-		} else if (st > SERD_FAILURE ||
-		           !read_PrefixedName(reader, *dest, false, &ate_dot) ||
-		           ate_dot) {
-			return (*dest = pop_node(reader, *dest));
-		} else {
-			return true;
-		}
 	}
-	return false;
+
+	/* Either a qname, or "a".  Read the prefix first, and if it is in fact
+	   "a", produce that instead.
+	*/
+	*dest = push_node(reader, SERD_CURIE, "", 0);
+	SerdNode*        node    = deref(reader, *dest);
+	const SerdStatus st      = read_PN_PREFIX(reader, *dest);
+	bool             ate_dot = false;
+	if (!st && node->n_bytes == 1 && node->buf[0] == 'a' &&
+	    is_token_end(peek_byte(reader))) {
+		pop_node(reader, *dest);
+		return (*dest = push_node(reader, SERD_URI, NS_RDF "type", 47));
+	} else if (st > SERD_FAILURE ||
+	           !read_PrefixedName(reader, *dest, false, &ate_dot) ||
+	           ate_dot) {
+		return (*dest = pop_node(reader, *dest));
+	}
+
+	return true;
 }
 
 static Ref
@@ -1279,18 +1259,16 @@ read_objectList(SerdReader* reader, ReadContext ctx, bool* ate_dot)
 static bool
 read_predicateObjectList(SerdReader* reader, ReadContext ctx, bool* ate_dot)
 {
-	uint8_t c;
-	while (true) {
-		TRY_THROW(read_verb(reader, &ctx.predicate));
-		read_ws_star(reader);
-
-		TRY_THROW(read_objectList(reader, ctx, ate_dot));
+	while (read_verb(reader, &ctx.predicate) &&
+	       read_ws_star(reader) &&
+	       read_objectList(reader, ctx, ate_dot)) {
 		ctx.predicate = pop_node(reader, ctx.predicate);
 		if (*ate_dot) {
 			return true;
 		}
 
-		bool ate_semi = false;
+		bool    ate_semi = false;
+		uint8_t c;
 		do {
 			read_ws_star(reader);
 			switch (c = peek_byte(reader)) {
@@ -1309,11 +1287,7 @@ read_predicateObjectList(SerdReader* reader, ReadContext ctx, bool* ate_dot)
 		}
 	}
 
-	pop_node(reader, ctx.predicate);
-	return true;
-except:
-	pop_node(reader, ctx.predicate);
-	return false;
+	return pop_node(reader, ctx.predicate);
 }
 
 static bool
@@ -1567,7 +1541,7 @@ tokcmp(SerdReader* reader, Ref ref, const char* tok, size_t n)
 			return ((*(uint8_t*)s1 < *(uint8_t*)s2) ? -1 : +1);
 		}
 	}
-    return 0;
+	return 0;
 }
 
 static bool
@@ -1583,7 +1557,7 @@ read_statement(SerdReader* reader)
 	switch (peek_byte(reader)) {
 	case '\0':
 		reader->eof = true;
-		return !reader->error;
+		return reader->status <= SERD_FAILURE;
 	case '@':
 		TRY_RET(read_directive(reader));
 		read_ws_star(reader);
@@ -1636,7 +1610,7 @@ read_turtleDoc(SerdReader* reader)
 	while (!reader->eof) {
 		TRY_RET(read_statement(reader));
 	}
-	return !reader->error;
+	return reader->status <= SERD_FAILURE;
 }
 
 static bool
@@ -1645,7 +1619,7 @@ read_trigDoc(SerdReader* reader)
 	while (!reader->eof) {
 		TRY_RET(read_statement(reader));
 	}
-	return !reader->error;
+	return reader->status <= SERD_FAILURE;
 }
 
 static bool
@@ -1696,7 +1670,7 @@ read_nquadsDoc(SerdReader* reader)
 		pop_node(reader, ctx.datatype);
 		pop_node(reader, ctx.object);
 	}
-	return !reader->error;
+	return reader->status <= SERD_FAILURE;
 }
 
 static bool
@@ -1830,104 +1804,143 @@ serd_reader_read_file(SerdReader*    reader,
 	return ret;
 }
 
-static void
+static SerdStatus
 skip_bom(SerdReader* me)
 {
-	const uint8_t* const b = me->read_buf;
-	if (me->paging && b[0] == 0xEF && b[1] == 0xBB && b[2] == 0xBF) {
-		me->read_head += 3;
+	if (peek_byte(me) == 0xEF) {
+		eat_byte_safe(me, 0xEF);
+		if (eat_byte_check(me, 0xBB) != 0xBB ||
+		    eat_byte_check(me, 0xBF) != 0xBF) {
+			r_err(me, SERD_ERR_BAD_SYNTAX, "corrupt byte order mark\n");
+			return SERD_ERR_BAD_SYNTAX;
+		}
 	}
+
+	return SERD_SUCCESS;
 }
 
 SERD_API
 SerdStatus
-serd_reader_start_stream(SerdReader*    me,
+serd_reader_start_stream(SerdReader*    reader,
                          FILE*          file,
                          const uint8_t* name,
                          bool           bulk)
 {
-	const Cursor cur = { name, 1, 1 };
-	me->fd        = file;
-	me->read_head = 0;
-	me->cur       = cur;
-	me->from_file = true;
-	me->eof       = false;
-	me->paging    = bulk;
+	return serd_reader_start_source_stream(
+		reader,
+		bulk ? (SerdSource)fread : serd_file_read_byte,
+		(SerdStreamErrorFunc)ferror,
+		file,
+		name,
+		bulk ? SERD_PAGE_SIZE : 1);
+}
 
-	if (bulk) {
-		me->file_buf = (uint8_t*)serd_bufalloc(SERD_PAGE_SIZE);
-		me->read_buf = me->file_buf;
-		memset(me->file_buf, '\0', SERD_PAGE_SIZE);
-		SerdStatus st = page(me);
-		if (st) {
-			serd_reader_end_stream(me);
+SERD_API
+SerdStatus
+serd_reader_start_source_stream(SerdReader*         reader,
+                                SerdSource          read_func,
+                                SerdStreamErrorFunc error_func,
+                                void*               stream,
+                                const uint8_t*      name,
+                                size_t              page_size)
+{
+	const Cursor cur = { name, 1, 1 };
+	reader->cur = cur;
+
+	return serd_byte_source_open_source(
+		&reader->source, read_func, error_func, stream, page_size);
+}
+
+static SerdStatus
+serd_reader_prepare(SerdReader* reader)
+{
+	reader->eof    = false;
+	reader->status = serd_byte_source_prepare(&reader->source);
+	if (reader->status == SERD_SUCCESS) {
+		reader->status = skip_bom(reader);
+	} else if (reader->status == SERD_FAILURE) {
+		reader->eof = true;
+	} else {
+		r_err(reader, reader->status, "read error: %s\n", strerror(errno));
+	}
+	return reader->status;
+}
+
+SERD_API
+SerdStatus
+serd_reader_read_chunk(SerdReader* reader)
+{
+	SerdStatus st = SERD_SUCCESS;
+	if (!reader->source.prepared) {
+		if ((st = serd_reader_prepare(reader))) {
 			return st;
 		}
-		skip_bom(me);
-	} else {
-		me->read_buf  = &me->read_byte;
-		me->read_byte = 0;  // Don't read to avoid potentially blocking
-	}
-
-	return SERD_SUCCESS;
-}
-
-SERD_API
-SerdStatus
-serd_reader_read_chunk(SerdReader* me)
-{
-	if (!me->read_byte) {
-		// Read initial byte
-		const int c = fgetc(me->fd);
-		me->read_byte = (c == EOF) ? 0 : (uint8_t)c;
-		if (c == EOF) {
-			me->eof = true;
-			return SERD_FAILURE;
+	} else if (reader->eof) {
+		reader->eof = false;
+		if ((st = serd_byte_source_advance(&reader->source))) {
+			return st;
 		}
 	}
-	return read_statement(me) ? SERD_SUCCESS : SERD_FAILURE;
+	return read_statement(reader) ? SERD_SUCCESS : SERD_FAILURE;
 }
 
 SERD_API
 SerdStatus
-serd_reader_end_stream(SerdReader* me)
+serd_reader_end_stream(SerdReader* reader)
 {
-	if (me->paging) {
-		free(me->file_buf);
-	}
-	me->fd       = NULL;
-	me->read_buf = me->file_buf = NULL;
-	return SERD_SUCCESS;
+	return serd_byte_source_close(&reader->source);
 }
 
 SERD_API
 SerdStatus
-serd_reader_read_file_handle(SerdReader* me, FILE* file, const uint8_t* name)
+serd_reader_read_file_handle(SerdReader*    reader,
+                             FILE*          file,
+                             const uint8_t* name)
 {
-	SerdStatus st = serd_reader_start_stream(me, file, name, true);
-	if (!st) {
-		st = read_doc(me) ? SERD_SUCCESS : SERD_ERR_UNKNOWN;
-		serd_reader_end_stream(me);
-	}
-	return st;
+	return serd_reader_read_source(
+		reader, (SerdSource)fread, (SerdStreamErrorFunc)ferror,
+		file, name, SERD_PAGE_SIZE);
 }
 
 SERD_API
 SerdStatus
-serd_reader_read_string(SerdReader* me, const uint8_t* utf8)
+serd_reader_read_source(SerdReader*         reader,
+                        SerdSource          source,
+                        SerdStreamErrorFunc error,
+                        void*               stream,
+                        const uint8_t*      name,
+                        size_t              page_size)
+{
+	SerdStatus st = serd_reader_start_source_stream(
+		reader, source, error, stream, name, page_size);
+
+	if (st || (st = serd_reader_prepare(reader))) {
+		serd_reader_end_stream(reader);
+		return st;
+	} else if (!read_doc(reader)) {
+		serd_reader_end_stream(reader);
+		return SERD_ERR_UNKNOWN;
+	}
+
+	return serd_reader_end_stream(reader);
+}
+
+SERD_API
+SerdStatus
+serd_reader_read_string(SerdReader* reader, const uint8_t* utf8)
 {
 	const Cursor cur = { (const uint8_t*)"(string)", 1, 1 };
 
-	me->read_buf  = utf8;
-	me->read_head = 0;
-	me->cur       = cur;
-	me->from_file = false;
-	me->paging    = false;
-	me->eof       = false;
+	serd_byte_source_open_string(&reader->source, utf8);
+	reader->cur = cur;
+	reader->eof = false;
 
-	skip_bom(me);
-	const bool ret = read_doc(me);
+	SerdStatus st = serd_reader_prepare(reader);
+	if (!st) {
+		st = read_doc(reader) ? SERD_SUCCESS : SERD_ERR_UNKNOWN;
+	}
 
-	me->read_buf = NULL;
-	return ret ? SERD_SUCCESS : SERD_ERR_UNKNOWN;
+	serd_byte_source_close(&reader->source);
+
+	return st;
 }
