@@ -43,7 +43,7 @@
 static inline FILE*
 serd_fopen(const char* path, const char* mode)
 {
-	FILE* fd = fopen((const char*)path, mode);
+	FILE* fd = fopen(path, mode);
 	if (!fd) {
 		fprintf(stderr, "Error opening file %s (%s)\n", path, strerror(errno));
 		return NULL;
@@ -66,6 +66,52 @@ serd_bufalloc(size_t size)
 #endif
 }
 
+/* Byte source */
+
+typedef struct {
+	SerdSource          read_func;    ///< Read function (e.g. fread)
+	SerdStreamErrorFunc error_func;   ///< Error function (e.g. ferror)
+	void*               stream;       ///< Stream (e.g. FILE)
+	size_t              page_size;    ///< Number of bytes to read at a time
+	uint8_t*            file_buf;     ///< Buffer iff reading pages from a file
+	const uint8_t*      read_buf;     ///< Pointer to file_buf or read_byte
+	size_t              read_head;    ///< Offset into read_buf
+	uint8_t             read_byte;    ///< 1-byte 'buffer' used when not paging
+	bool                from_stream;  ///< True iff reading from `stream`
+	bool                prepared;     ///< True iff prepared for reading
+} SerdByteSource;
+
+SerdStatus
+serd_byte_source_open_file(SerdByteSource* source,
+                           FILE*           file,
+                           bool            bulk);
+
+SerdStatus
+serd_byte_source_open_string(SerdByteSource* source, const uint8_t* utf8);
+
+SerdStatus
+serd_byte_source_open_source(SerdByteSource*     source,
+                             SerdSource          read_func,
+                             SerdStreamErrorFunc error_func,
+                             void*               stream,
+                             size_t              page_size);
+
+SerdStatus
+serd_byte_source_close(SerdByteSource* source);
+
+SerdStatus
+serd_byte_source_prepare(SerdByteSource* source);
+
+static inline uint8_t
+serd_byte_source_peek(SerdByteSource* source)
+{
+	assert(source->prepared);
+	return source->read_buf[source->read_head];
+}
+
+SerdStatus
+serd_byte_source_advance(SerdByteSource* source);
+
 /* Stack */
 
 /** A dynamic stack in memory. */
@@ -155,48 +201,56 @@ serd_stack_pop_aligned(SerdStack* stack, size_t n_bytes)
 	serd_stack_pop(stack, pad + 1);
 }
 
-/* Bulk Sink */
+/* Byte Sink */
 
-typedef struct SerdBulkSinkImpl {
+typedef struct SerdByteSinkImpl {
 	SerdSink sink;
 	void*    stream;
 	uint8_t* buf;
 	size_t   size;
 	size_t   block_size;
-} SerdBulkSink;
+} SerdByteSink;
 
-static inline SerdBulkSink
-serd_bulk_sink_new(SerdSink sink, void* stream, size_t block_size)
+static inline SerdByteSink
+serd_byte_sink_new(SerdSink sink, void* stream, size_t block_size)
 {
-	SerdBulkSink bsink;
+	SerdByteSink bsink;
 	bsink.sink       = sink;
 	bsink.stream     = stream;
 	bsink.size       = 0;
 	bsink.block_size = block_size;
-	bsink.buf        = (uint8_t*)serd_bufalloc(block_size);
+	bsink.buf        = ((block_size > 1)
+	                    ? (uint8_t*)serd_bufalloc(block_size)
+	                    : NULL);
 	return bsink;
 }
 
 static inline void
-serd_bulk_sink_flush(SerdBulkSink* bsink)
+serd_byte_sink_flush(SerdByteSink* bsink)
 {
-	if (bsink->size > 0) {
+	if (bsink->block_size > 1 && bsink->size > 0) {
 		bsink->sink(bsink->buf, bsink->size, bsink->stream);
+		bsink->size = 0;
 	}
-	bsink->size = 0;
 }
 
 static inline void
-serd_bulk_sink_free(SerdBulkSink* bsink)
+serd_byte_sink_free(SerdByteSink* bsink)
 {
-	serd_bulk_sink_flush(bsink);
+	serd_byte_sink_flush(bsink);
 	free(bsink->buf);
 	bsink->buf = NULL;
 }
 
 static inline size_t
-serd_bulk_sink_write(const void* buf, size_t len, SerdBulkSink* bsink)
+serd_byte_sink_write(const void* buf, size_t len, SerdByteSink* bsink)
 {
+	if (len == 0) {
+		return 0;
+	} else if (bsink->block_size == 1) {
+		return bsink->sink(buf, len, bsink->stream);
+	}
+
 	const size_t orig_len = len;
 	while (len) {
 		const size_t space = bsink->block_size - bsink->size;
@@ -240,6 +294,13 @@ is_digit(const uint8_t c)
 	return in_range(c, '0', '9');
 }
 
+/* RFC2234: HEXDIG ::= DIGIT / "A" / "B" / "C" / "D" / "E" / "F" */
+static inline bool
+is_hexdig(const uint8_t c)
+{
+	return is_digit(c) || in_range(c, 'A', 'F');
+}
+
 static inline bool
 is_space(const char c)
 {
@@ -264,6 +325,35 @@ is_windows_path(const uint8_t* path)
 		&& (path[2] == '/' || path[2] == '\\');
 }
 
+/* String utilities */
+
+size_t
+serd_substrlen(const uint8_t* str,
+               const size_t   len,
+               size_t*        n_bytes,
+               SerdNodeFlags* flags);
+
+static inline uint32_t
+utf8_num_bytes(const uint8_t c)
+{
+	if ((c & 0x80) == 0) {  // Starts with `0'
+		return 1;
+	}
+
+#ifdef HAVE_BUILTIN_CLZ
+	return __builtin_clz(~c << 24);
+#else
+	if ((c & 0xE0) == 0xC0) {  // Starts with `110'
+		return 2;
+	} else if ((c & 0xF0) == 0xE0) {  // Starts with `1110'
+		return 3;
+	} else if ((c & 0xF8) == 0xF0) {  // Starts with `11110'
+		return 4;
+	}
+	return 0;
+#endif
+}
+
 /* URI utilities */
 
 static inline bool
@@ -314,6 +404,17 @@ uri_is_under(const SerdURI* uri, const SerdURI* root)
 	return true;
 }
 
+static inline bool
+is_uri_scheme_char(const uint8_t c)
+{
+	switch (c) {
+	case ':': case '+': case '-': case '.':
+		return true;
+	default:
+		return is_alpha(c) || is_digit(c);
+	}
+}
+
 /* Error reporting */
 
 static inline void
