@@ -1,5 +1,5 @@
 /*
-  Copyright 2011-2012 David Robillard <http://drobilla.net>
+  Copyright 2011-2016 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -20,9 +20,9 @@
 
 SERD_API
 const uint8_t*
-serd_strerror(SerdStatus st)
+serd_strerror(SerdStatus status)
 {
-	switch (st) {
+	switch (status) {
 	case SERD_SUCCESS:        return (const uint8_t*)"Success";
 	case SERD_FAILURE:        return (const uint8_t*)"Non-fatal failure";
 	case SERD_ERR_UNKNOWN:    return (const uint8_t*)"Unknown error";
@@ -36,29 +36,61 @@ serd_strerror(SerdStatus st)
 	return (const uint8_t*)"Unknown error";  // never reached
 }
 
-SERD_API
-size_t
-serd_strlen(const uint8_t* str, size_t* n_bytes, SerdNodeFlags* flags)
+static inline void
+serd_update_flags(const uint8_t c, SerdNodeFlags* const flags)
 {
-	size_t n_chars = 0;
-	size_t i       = 0;
-	*flags         = 0;
-	for (; str[i]; ++i) {
-		if ((str[i] & 0xC0) != 0x80) {
-			// Does not start with `10', start of a new character
+	switch (c) {
+	case '\r': case '\n':
+		*flags |= SERD_HAS_NEWLINE;
+		break;
+	case '"':
+		*flags |= SERD_HAS_QUOTE;
+	}
+}
+
+size_t
+serd_substrlen(const uint8_t* const str,
+               const size_t         len,
+               size_t* const        n_bytes,
+               SerdNodeFlags* const flags)
+{
+	size_t        n_chars = 0;
+	size_t        i       = 0;
+	SerdNodeFlags f       = 0;
+	for (; i < len && str[i]; ++i) {
+		if ((str[i] & 0xC0) != 0x80) {  // Start of new character
 			++n_chars;
-			switch (str[i]) {
-			case '\r': case '\n':
-				*flags |= SERD_HAS_NEWLINE;
-				break;
-			case '"':
-				*flags |= SERD_HAS_QUOTE;
-			}
+			serd_update_flags(str[i], &f);
 		}
 	}
 	if (n_bytes) {
 		*n_bytes = i;
 	}
+	if (flags) {
+		*flags = f;
+	}
+	return n_chars;
+}
+
+SERD_API
+size_t
+serd_strlen(const uint8_t* str, size_t* n_bytes, SerdNodeFlags* flags)
+{
+	size_t        n_chars = 0;
+	size_t        i       = 0;
+	SerdNodeFlags f       = 0;
+	for (; str[i]; ++i) {
+		if ((str[i] & 0xC0) != 0x80) {  // Start of new character
+			++n_chars;
+			serd_update_flags(str[i], &f);
+		}
+	}
+	if (n_bytes) {
+		*n_bytes = i;
+	}
+	if (flags) {
+		*flags = f;
+	}
 	return n_chars;
 }
 
