@@ -1,5 +1,5 @@
 /*
-  Copyright 2011-2012 David Robillard <http://drobilla.net>
+  Copyright 2011-2014 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -78,7 +78,7 @@ serd_file_uri_parse(const uint8_t* uri, uint8_t** hostname)
 			if (*(s + 1) == '%') {
 				serd_chunk_sink("%", 1, &chunk);
 				++s;
-			} else if (is_digit(*(s + 1)) && is_digit(*(s + 2))) {
+			} else if (is_hexdig(*(s + 1)) && is_hexdig(*(s + 2))) {
 				const uint8_t code[3] = { *(s + 1), *(s + 2), 0 };
 				uint32_t num;
 				sscanf((const char*)code, "%X", &num);
@@ -100,19 +100,15 @@ bool
 serd_uri_string_has_scheme(const uint8_t* utf8)
 {
 	// RFC3986: scheme ::= ALPHA *( ALPHA / DIGIT / "+" / "-" / "." )
-	if (!is_alpha(utf8[0])) {
+	if (!utf8 || !is_alpha(utf8[0])) {
 		return false;  // Invalid scheme initial character, URI is relative
 	}
+
 	for (uint8_t c; (c = *++utf8) != '\0';) {
-		switch (c) {
-		case ':':
+		if (!is_uri_scheme_char(c)) {
+			return false;
+		} else if (c == ':') {
 			return true;  // End of scheme
-		case '+': case '-': case '.':
-			break;  // Valid scheme character, continue
-		default:
-			if (!is_alpha(c) && !is_digit(c)) {
-				return false;  // Invalid scheme character
-			}
 		}
 	}
 
@@ -141,9 +137,9 @@ serd_uri_dump(const SerdURI* uri, FILE* file)
 
 SERD_API
 SerdStatus
-serd_uri_parse(const uint8_t* utf8, SerdURI* uri)
+serd_uri_parse(const uint8_t* utf8, SerdURI* out)
 {
-	*uri = SERD_URI_NULL;
+	*out = SERD_URI_NULL;
 
 	const uint8_t* ptr = utf8;
 
@@ -159,8 +155,8 @@ serd_uri_parse(const uint8_t* utf8, SerdURI* uri)
 				ptr = utf8;
 				goto path;  // Relative URI (starts with path by definition)
 			case ':':
-				uri->scheme.buf = utf8;
-				uri->scheme.len = (ptr++) - utf8;
+				out->scheme.buf = utf8;
+				out->scheme.len = (ptr++) - utf8;
 				goto maybe_authority;  // URI with scheme
 			case '+': case '-': case '.':
 				continue;
@@ -179,14 +175,14 @@ serd_uri_parse(const uint8_t* utf8, SerdURI* uri)
 maybe_authority:
 	if (*ptr == '/' && *(ptr + 1) == '/') {
 		ptr += 2;
-		uri->authority.buf = ptr;
+		out->authority.buf = ptr;
 		for (uint8_t c; (c = *ptr) != '\0'; ++ptr) {
 			switch (c) {
 			case '/': goto path;
 			case '?': goto query;
 			case '#': goto fragment;
 			default:
-				++uri->authority.len;
+				++out->authority.len;
 			}
 		}
 	}
@@ -201,14 +197,14 @@ path:
 	case '\0': goto end;
 	default:  break;
 	}
-	uri->path.buf = ptr;
-	uri->path.len = 0;
+	out->path.buf = ptr;
+	out->path.len = 0;
 	for (uint8_t c; (c = *ptr) != '\0'; ++ptr) {
 		switch (c) {
 		case '?': goto query;
 		case '#': goto fragment;
 		default:
-			++uri->path.len;
+			++out->path.len;
 		}
 	}
 
@@ -218,13 +214,13 @@ path:
 	*/
 query:
 	if (*ptr == '?') {
-		uri->query.buf = ++ptr;
+		out->query.buf = ++ptr;
 		for (uint8_t c; (c = *ptr) != '\0'; ++ptr) {
 			switch (c) {
 			case '#':
 				goto fragment;
 			default:
-				++uri->query.len;
+				++out->query.len;
 			}
 		}
 	}
@@ -235,16 +231,16 @@ query:
 	*/
 fragment:
 	if (*ptr == '#') {
-		uri->fragment.buf = ptr;
+		out->fragment.buf = ptr;
 		while (*ptr++ != '\0') {
-			++uri->fragment.len;
+			++out->fragment.len;
 		}
 	}
 
 end:
 	#ifdef URI_DEBUG
 	fprintf(stderr, "PARSE URI <%s>\n", utf8);
-	serd_uri_dump(uri, stderr);
+	serd_uri_dump(out, stderr);
 	fprintf(stderr, "\n");
 	#endif
 
@@ -252,10 +248,10 @@ end:
 }
 
 /**
-   Remove leading dot components from @c path.
+   Remove leading dot components from `path`.
    See http://tools.ietf.org/html/rfc3986#section-5.2.3
    @param up Set to the number of up-references (e.g. "../") trimmed
-   @return A pointer to the new start of @path
+   @return A pointer to the new start of `path`
 */
 static const uint8_t*
 remove_dot_segments(const uint8_t* path, size_t len, size_t* up)
@@ -317,7 +313,7 @@ remove_dot_segments(const uint8_t* path, size_t len, size_t* up)
 	return begin;
 }
 
-/// Merge @p base and @p path in-place
+/// Merge `base` and `path` in-place
 static void
 merge(SerdChunk* base, SerdChunk* path)
 {
@@ -325,7 +321,7 @@ merge(SerdChunk* base, SerdChunk* path)
 	const uint8_t* begin = remove_dot_segments(path->buf, path->len, &up);
 	const uint8_t* end   = path->buf + path->len;
 
-	if (base->buf && base->len > 0) {
+	if (base->len) {
 		// Find the up'th last slash
 		const uint8_t* base_last = (base->buf + base->len - 1);
 		++up;
@@ -396,7 +392,7 @@ serd_uri_resolve(const SerdURI* r, const SerdURI* base, SerdURI* t)
 	#endif
 }
 
-/** Write the path of @p uri starting at index @p i */
+/** Write the path of `uri` starting at index `i` */
 static size_t
 write_path_tail(SerdSink sink, void* stream, const SerdURI* uri, size_t i)
 {
@@ -415,7 +411,7 @@ write_path_tail(SerdSink sink, void* stream, const SerdURI* uri, size_t i)
 	return len;
 }
 
-/** Write the path of @p uri relative to the path of @p base. */
+/** Write the path of `uri` relative to the path of `base`. */
 static size_t
 write_rel_path(SerdSink       sink,
                void*          stream,
