@@ -49,7 +49,9 @@ typedef enum {
 	SEP_LIST_SEP,    ///< List separator (whitespace)
 	SEP_LIST_END,    ///< End of list (')')
 	SEP_GRAPH_BEGIN, ///< Start of graph ('{')
-	SEP_GRAPH_END    ///< End of graph ('}')
+	SEP_GRAPH_END,   ///< End of graph ('}')
+	SEP_URI_BEGIN,   ///< URI start quote ('<')
+	SEP_URI_END      ///< URI end quote ('>')
 } Sep;
 
 typedef struct {
@@ -60,6 +62,8 @@ typedef struct {
 	uint8_t     space_after_sep;   ///< Newline after sep if after sep
 } SepRule;
 
+static const uint8_t replacement_char[] = { 0xEF, 0xBF, 0xBD };
+
 static const SepRule rules[] = {
 	{ NULL,     0, 0, 0, 0 },
 	{ " .\n\n", 4, 0, 0, 0 },
@@ -74,6 +78,8 @@ static const SepRule rules[] = {
 	{ ")",      1, 1, 0, 0 },
 	{ " {",     2, 0, 1, 1 },
 	{ " }",     2, 0, 1, 1 },
+	{ "<",      1, 0, 0, 0 },
+	{ ">",      1, 0, 0, 0 },
 	{ "\n",     1, 0, 1, 0 }
 };
 
@@ -85,17 +91,15 @@ struct SerdWriterImpl {
 	SerdURI       root_uri;
 	SerdURI       base_uri;
 	SerdStack     anon_stack;
-	SerdBulkSink  bulk_sink;
-	SerdSink      sink;
-	void*         stream;
+	SerdByteSink  byte_sink;
 	SerdErrorSink error_sink;
 	void*         error_handle;
 	WriteContext  context;
 	SerdNode      list_subj;
 	unsigned      list_depth;
+	unsigned      indent;
 	uint8_t*      bprefix;
 	size_t        bprefix_len;
-	unsigned      indent;
 	Sep           last_sep;
 	bool          empty;
 };
@@ -105,6 +109,22 @@ typedef enum {
 	WRITE_LONG_STRING
 } TextContext;
 
+typedef enum {
+	FIELD_NONE,
+	FIELD_SUBJECT,
+	FIELD_PREDICATE,
+	FIELD_OBJECT,
+	FIELD_GRAPH
+} Field;
+
+static bool
+write_node(SerdWriter*        writer,
+           const SerdNode*    node,
+           const SerdNode*    datatype,
+           const SerdNode*    lang,
+           Field              field,
+           SerdStatementFlags flags);
+
 static bool
 supports_abbrev(const SerdWriter* writer)
 {
@@ -153,51 +173,25 @@ copy_node(SerdNode* dst, const SerdNode* src)
 static inline size_t
 sink(const void* buf, size_t len, SerdWriter* writer)
 {
-	if (len == 0) {
-		return 0;
-	} else if (writer->style & SERD_STYLE_BULK) {
-		return serd_bulk_sink_write(buf, len, &writer->bulk_sink);
-	} else {
-		return writer->sink(buf, len, writer->stream);
-	}
+	return serd_byte_sink_write(buf, len, &writer->byte_sink);
 }
 
 // Parse a UTF-8 character, set *size to the length, and return the code point
 static inline uint32_t
 parse_utf8_char(SerdWriter* writer, const uint8_t* utf8, size_t* size)
 {
-	uint32_t c = 0;
-	if ((utf8[0] & 0x80) == 0) {  // Starts with `0'
-		*size = 1;
-		c     = utf8[0];
-	} else if ((utf8[0] & 0xE0) == 0xC0) {  // Starts with `110'
-		*size = 2;
-		c     = utf8[0] & 0x1F;
-	} else if ((utf8[0] & 0xF0) == 0xE0) {  // Starts with `1110'
-		*size = 3;
-		c     = utf8[0] & 0x0F;
-	} else if ((utf8[0] & 0xF8) == 0xF0) {  // Starts with `11110'
-		*size = 4;
-		c     = utf8[0] & 0x07;
-	} else {
-		w_err(writer, SERD_ERR_BAD_ARG, "invalid UTF-8: %X\n", utf8[0]);
-		*size = 0;
-		return 0;
+	switch (*size = utf8_num_bytes(utf8[0])) {
+	case 1: case 2: case 3: case 4:
+		break;
+	default:
+		return *size = 0;
 	}
 
-	size_t  i  = 0;
-	uint8_t in = utf8[i++];
-
-#define READ_BYTE() \
-	in = utf8[i++] & 0x3F; \
-	c  = (c << 6) | in;
-
-	switch (*size) {
-	case 4: READ_BYTE();
-	case 3: READ_BYTE();
-	case 2: READ_BYTE();
+	uint32_t c = utf8[0] & ((1 << (8 - *size)) - 1);
+	for (size_t i = 1; i < *size; ++i) {
+		const uint8_t in = utf8[i] & 0x3F;
+		c = (c << 6) | in;
 	}
-
 	return c;
 }
 
@@ -206,17 +200,14 @@ parse_utf8_char(SerdWriter* writer, const uint8_t* utf8, size_t* size)
 static size_t
 write_character(SerdWriter* writer, const uint8_t* utf8, size_t* size)
 {
-	const uint8_t replacement_char[] = { 0xEF, 0xBF, 0xBD };
-	char          escape[11]         = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
-	const uint8_t in                 = utf8[0];
-
-	uint32_t c = parse_utf8_char(writer, utf8, size);
+	char           escape[11] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
+	const uint32_t c          = parse_utf8_char(writer, utf8, size);
 	switch (*size) {
 	case 0:
-		w_err(writer, SERD_ERR_BAD_ARG, "invalid UTF-8: %X\n", in);
+		w_err(writer, SERD_ERR_BAD_ARG, "invalid UTF-8: %X\n", utf8[0]);
 		return sink(replacement_char, sizeof(replacement_char), writer);
 	case 1:
-		snprintf(escape, sizeof(escape), "\\u%04X", in);
+		snprintf(escape, sizeof(escape), "\\u%04X", utf8[0]);
 		return sink(escape, 6, writer);
 	default:
 		break;
@@ -270,6 +261,11 @@ write_uri(SerdWriter* writer, const uint8_t* utf8, size_t n_bytes)
 		size_t size = 0;
 		len += write_character(writer, utf8 + i, &size);
 		i   += size;
+		if (size == 0) {
+			// Corrupt input, write replacement char and scan to next start
+			sink(replacement_char, sizeof(replacement_char), writer);
+			for (; i < n_bytes && (utf8[i] & 0x80); ++i) {}
+		}
 	}
 	return len;
 }
@@ -376,9 +372,10 @@ write_text(SerdWriter* writer, TextContext ctx,
 
 		size_t size = 0;
 		len += write_character(writer, utf8 + i - 1, &size);
-
 		if (size == 0) {
-			return len;
+			// Corrupt input, write replacement char and scan to next start
+			sink(replacement_char, sizeof(replacement_char), writer);
+			for (; i < n_bytes && (utf8[i] & 0x80); ++i) {}
 		}
 
 		i += size - 1;
@@ -401,7 +398,7 @@ write_newline(SerdWriter* writer)
 	}
 }
 
-static void
+static bool
 write_sep(SerdWriter* writer, const Sep sep)
 {
 	const SepRule* rule = &rules[sep];
@@ -411,13 +408,14 @@ write_sep(SerdWriter* writer, const Sep sep)
 	if (rule->str) {
 		sink(rule->str, rule->len, writer);
 	}
-	if (    (writer->last_sep && rule->space_after_sep)
-	    || (!writer->last_sep && rule->space_after_node)) {
+	if ((writer->last_sep && rule->space_after_sep) ||
+	    (!writer->last_sep && rule->space_after_node)) {
 		write_newline(writer);
 	} else if (writer->last_sep && rule->space_after_node) {
 		sink(" ", 1, writer);
 	}
 	writer->last_sep = sep;
+	return true;
 }
 
 static SerdStatus
@@ -441,14 +439,6 @@ free_context(SerdWriter* writer)
 	return reset_context(writer, true);
 }
 
-typedef enum {
-	FIELD_NONE,
-	FIELD_SUBJECT,
-	FIELD_PREDICATE,
-	FIELD_OBJECT,
-	FIELD_GRAPH
-} Field;
-
 static bool
 is_inline_start(const SerdWriter* writer, Field field, SerdStatementFlags flags)
 {
@@ -457,6 +447,188 @@ is_inline_start(const SerdWriter* writer, Field field, SerdStatementFlags flags)
 	         (field == FIELD_OBJECT &&  (flags & SERD_ANON_O_BEGIN))));
 }
 
+static bool
+write_literal(SerdWriter*        writer,
+              const SerdNode*    node,
+              const SerdNode*    datatype,
+              const SerdNode*    lang,
+              Field              field,
+              SerdStatementFlags flags)
+{
+	if (supports_abbrev(writer) && datatype && datatype->buf) {
+		const char* type_uri = (const char*)datatype->buf;
+		if (!strncmp(type_uri, NS_XSD, sizeof(NS_XSD) - 1) && (
+			    !strcmp(type_uri + sizeof(NS_XSD) - 1, "boolean") ||
+			    !strcmp(type_uri + sizeof(NS_XSD) - 1, "integer"))) {
+			sink(node->buf, node->n_bytes, writer);
+			return true;
+		} else if (!strncmp(type_uri, NS_XSD, sizeof(NS_XSD) - 1) &&
+		           !strcmp(type_uri + sizeof(NS_XSD) - 1, "decimal") &&
+		           strchr((const char*)node->buf, '.') &&
+		           node->buf[node->n_bytes - 1] != '.') {
+			/* xsd:decimal literals without trailing digits, e.g. "5.", can
+			   not be written bare in Turtle.  We could add a 0 which is
+			   prettier, but changes the text and breaks round tripping.
+			*/
+			sink(node->buf, node->n_bytes, writer);
+			return true;
+		}
+	}
+
+	if (supports_abbrev(writer)
+	    && (node->flags & (SERD_HAS_NEWLINE|SERD_HAS_QUOTE))) {
+		sink("\"\"\"", 3, writer);
+		write_text(writer, WRITE_LONG_STRING, node->buf, node->n_bytes);
+		sink("\"\"\"", 3, writer);
+	} else {
+		sink("\"", 1, writer);
+		write_text(writer, WRITE_STRING, node->buf, node->n_bytes);
+		sink("\"", 1, writer);
+	}
+	if (lang && lang->buf) {
+		sink("@", 1, writer);
+		sink(lang->buf, lang->n_bytes, writer);
+	} else if (datatype && datatype->buf) {
+		sink("^^", 2, writer);
+		return write_node(writer, datatype, NULL, NULL, FIELD_NONE, flags);
+	}
+	return true;
+}
+
+static bool
+write_uri_node(SerdWriter* const        writer,
+               const SerdNode*          node,
+               const Field              field,
+               const SerdStatementFlags flags)
+{
+	SerdNode  prefix;
+	SerdChunk suffix;
+
+	if (is_inline_start(writer, field, flags)) {
+		++writer->indent;
+		write_sep(writer, SEP_ANON_BEGIN);
+		sink("== ", 3, writer);
+	}
+
+	const bool has_scheme = serd_uri_string_has_scheme(node->buf);
+	if (field == FIELD_PREDICATE && supports_abbrev(writer)
+	    && !strcmp((const char*)node->buf, NS_RDF "type")) {
+		return sink("a", 1, writer) == 1;
+	} else if (supports_abbrev(writer)
+	           && !strcmp((const char*)node->buf, NS_RDF "nil")) {
+		return sink("()", 2, writer) == 2;
+	} else if (has_scheme && (writer->style & SERD_STYLE_CURIED) &&
+	           serd_env_qualify(writer->env, node, &prefix, &suffix)) {
+		write_uri(writer, prefix.buf, prefix.n_bytes);
+		sink(":", 1, writer);
+		write_uri(writer, suffix.buf, suffix.len);
+		return true;
+	}
+
+	write_sep(writer, SEP_URI_BEGIN);
+	if (writer->style & SERD_STYLE_RESOLVED) {
+		SerdURI in_base_uri, uri, abs_uri;
+		serd_env_get_base_uri(writer->env, &in_base_uri);
+		serd_uri_parse(node->buf, &uri);
+		serd_uri_resolve(&uri, &in_base_uri, &abs_uri);
+		bool rooted = uri_is_under(&writer->base_uri, &writer->root_uri);
+		SerdURI* root = rooted ? &writer->root_uri : & writer->base_uri;
+		if (!uri_is_under(&abs_uri, root) ||
+		    writer->syntax == SERD_NTRIPLES ||
+		    writer->syntax == SERD_NQUADS) {
+			serd_uri_serialise(&abs_uri, uri_sink, writer);
+		} else {
+			serd_uri_serialise_relative(
+				&uri, &writer->base_uri, root, uri_sink, writer);
+		}
+	} else {
+		write_uri(writer, node->buf, node->n_bytes);
+	}
+	write_sep(writer, SEP_URI_END);
+	if (is_inline_start(writer, field, flags)) {
+		sink(" ;", 2, writer);
+		write_newline(writer);
+	}
+	return true;
+}
+
+static bool
+write_curie(SerdWriter* const        writer,
+            const SerdNode*          node,
+            const Field              field,
+            const SerdStatementFlags flags)
+{
+	SerdChunk prefix;
+	SerdChunk suffix;
+	switch (writer->syntax) {
+	case SERD_NTRIPLES:
+	case SERD_NQUADS:
+		if (serd_env_expand(writer->env, node, &prefix, &suffix)) {
+			w_err(writer, SERD_ERR_BAD_CURIE,
+			      "undefined namespace prefix `%s'\n", node->buf);
+			return false;
+		}
+		write_sep(writer, SEP_URI_BEGIN);
+		write_uri(writer, prefix.buf, prefix.len);
+		write_uri(writer, suffix.buf, suffix.len);
+		write_sep(writer, SEP_URI_END);
+		break;
+	case SERD_TURTLE:
+	case SERD_TRIG:
+		if (is_inline_start(writer, field, flags)) {
+			++writer->indent;
+			write_sep(writer, SEP_ANON_BEGIN);
+			sink("== ", 3, writer);
+		}
+		write_lname(writer, node->buf, node->n_bytes);
+		if (is_inline_start(writer, field, flags)) {
+			sink(" ;", 2, writer);
+			write_newline(writer);
+		}
+	}
+	return true;
+}
+
+static bool
+write_blank(SerdWriter* const        writer,
+            const SerdNode*          node,
+            const Field              field,
+            const SerdStatementFlags flags)
+{
+	if (supports_abbrev(writer)) {
+		if (is_inline_start(writer, field, flags)) {
+			++writer->indent;
+			return write_sep(writer, SEP_ANON_BEGIN);
+		} else if (field == FIELD_SUBJECT && (flags & SERD_LIST_S_BEGIN)) {
+			assert(writer->list_depth == 0);
+			copy_node(&writer->list_subj, node);
+			++writer->list_depth;
+			++writer->indent;
+			return write_sep(writer, SEP_LIST_BEGIN);
+		} else if (field == FIELD_OBJECT && (flags & SERD_LIST_O_BEGIN)) {
+			++writer->indent;
+			++writer->list_depth;
+			return write_sep(writer, SEP_LIST_BEGIN);
+		} else if ((field == FIELD_SUBJECT && (flags & SERD_EMPTY_S)) ||
+		           (field == FIELD_OBJECT && (flags & SERD_EMPTY_O))) {
+			return sink("[]", 2, writer) == 2;
+		}
+	}
+
+	sink("_:", 2, writer);
+	if (writer->bprefix && !strncmp((const char*)node->buf,
+	                                (const char*)writer->bprefix,
+	                                writer->bprefix_len)) {
+		sink(node->buf + writer->bprefix_len,
+		     node->n_bytes - writer->bprefix_len,
+		     writer);
+	} else {
+		sink(node->buf, node->n_bytes, writer);
+	}
+
+	return true;
+}
+
 static bool
 write_node(SerdWriter*        writer,
            const SerdNode*    node,
@@ -465,161 +637,23 @@ write_node(SerdWriter*        writer,
            Field              field,
            SerdStatementFlags flags)
 {
-	SerdChunk uri_prefix;
-	SerdNode  prefix;
-	SerdChunk suffix;
-	bool      has_scheme;
+	bool ret = false;
 	switch (node->type) {
-	case SERD_BLANK:
-		if (is_inline_start(writer, field, flags)) {
-			++writer->indent;
-			write_sep(writer, SEP_ANON_BEGIN);
-		} else if (supports_abbrev(writer)
-		           && (field == FIELD_SUBJECT && (flags & SERD_LIST_S_BEGIN))) {
-			assert(writer->list_depth == 0);
-			copy_node(&writer->list_subj, node);
-			++writer->list_depth;
-			++writer->indent;
-			write_sep(writer, SEP_LIST_BEGIN);
-		} else if (supports_abbrev(writer)
-		           && (field == FIELD_OBJECT && (flags & SERD_LIST_O_BEGIN))) {
-			++writer->indent;
-			++writer->list_depth;
-			write_sep(writer, SEP_LIST_BEGIN);
-		} else if (supports_abbrev(writer)
-		           && ((field == FIELD_SUBJECT && (flags & SERD_EMPTY_S))
-		               || (field == FIELD_OBJECT && (flags & SERD_EMPTY_O)))) {
-			sink("[]", 2, writer);
-		} else {
-			sink("_:", 2, writer);
-			if (writer->bprefix && !strncmp((const char*)node->buf,
-			                                (const char*)writer->bprefix,
-			                                writer->bprefix_len)) {
-				sink(node->buf + writer->bprefix_len,
-				     node->n_bytes - writer->bprefix_len,
-				     writer);
-			} else {
-				sink(node->buf, node->n_bytes, writer);
-			}
-		}
-		break;
-	case SERD_CURIE:
-		switch (writer->syntax) {
-		case SERD_NTRIPLES:
-		case SERD_NQUADS:
-			if (serd_env_expand(writer->env, node, &uri_prefix, &suffix)) {
-				w_err(writer, SERD_ERR_BAD_CURIE,
-				      "undefined namespace prefix `%s'\n", node->buf);
-				return false;
-			}
-			sink("<", 1, writer);
-			write_uri(writer, uri_prefix.buf, uri_prefix.len);
-			write_uri(writer, suffix.buf, suffix.len);
-			sink(">", 1, writer);
-			break;
-		case SERD_TURTLE:
-		case SERD_TRIG:
-			if (is_inline_start(writer, field, flags)) {
-				++writer->indent;
-				write_sep(writer, SEP_ANON_BEGIN);
-				sink("== ", 3, writer);
-			}
-			write_lname(writer, node->buf, node->n_bytes);
-			if (is_inline_start(writer, field, flags)) {
-				sink(" ;", 2, writer);
-				write_newline(writer);
-			}
-		}
-		break;
 	case SERD_LITERAL:
-		if (supports_abbrev(writer) && datatype && datatype->buf) {
-			const char* type_uri = (const char*)datatype->buf;
-			if (!strncmp(type_uri, NS_XSD, sizeof(NS_XSD) - 1) && (
-				    !strcmp(type_uri + sizeof(NS_XSD) - 1, "boolean") ||
-				    !strcmp(type_uri + sizeof(NS_XSD) - 1, "integer"))) {
-				sink(node->buf, node->n_bytes, writer);
-				break;
-			} else if (!strncmp(type_uri, NS_XSD, sizeof(NS_XSD) - 1) &&
-			           !strcmp(type_uri + sizeof(NS_XSD) - 1, "decimal") &&
-			           strchr((const char*)node->buf, '.') &&
-			           node->buf[node->n_bytes - 1] != '.') {
-				/* xsd:decimal literals without trailing digits, e.g. "5.", can
-				   not be written bare in Turtle.  We could add a 0 which is
-				   prettier, but changes the text and breaks round tripping.
-				*/
-				sink(node->buf, node->n_bytes, writer);
-				break;
-			}
-		}
-		if (supports_abbrev(writer)
-		    && (node->flags & (SERD_HAS_NEWLINE|SERD_HAS_QUOTE))) {
-			sink("\"\"\"", 3, writer);
-			write_text(writer, WRITE_LONG_STRING, node->buf, node->n_bytes);
-			sink("\"\"\"", 3, writer);
-		} else {
-			sink("\"", 1, writer);
-			write_text(writer, WRITE_STRING, node->buf, node->n_bytes);
-			sink("\"", 1, writer);
-		}
-		if (lang && lang->buf) {
-			sink("@", 1, writer);
-			sink(lang->buf, lang->n_bytes, writer);
-		} else if (datatype && datatype->buf) {
-			sink("^^", 2, writer);
-			write_node(writer, datatype, NULL, NULL, FIELD_NONE, flags);
-		}
+		ret = write_literal(writer, node, datatype, lang, field, flags);
 		break;
 	case SERD_URI:
-		if (is_inline_start(writer, field, flags)) {
-			++writer->indent;
-			write_sep(writer, SEP_ANON_BEGIN);
-			sink("== ", 3, writer);
-		}
-		has_scheme = serd_uri_string_has_scheme(node->buf);
-		if (field == FIELD_PREDICATE && supports_abbrev(writer)
-		    && !strcmp((const char*)node->buf, NS_RDF "type")) {
-			sink("a", 1, writer);
-			break;
-		} else if (supports_abbrev(writer)
-		           && !strcmp((const char*)node->buf, NS_RDF "nil")) {
-			sink("()", 2, writer);
-			break;
-		} else if (has_scheme && (writer->style & SERD_STYLE_CURIED) &&
-		           serd_env_qualify(writer->env, node, &prefix, &suffix)) {
-			write_uri(writer, prefix.buf, prefix.n_bytes);
-			sink(":", 1, writer);
-			write_uri(writer, suffix.buf, suffix.len);
-			break;
-		}
-		sink("<", 1, writer);
-		if (writer->style & SERD_STYLE_RESOLVED) {
-			SerdURI in_base_uri, uri, abs_uri;
-			serd_env_get_base_uri(writer->env, &in_base_uri);
-			serd_uri_parse(node->buf, &uri);
-			serd_uri_resolve(&uri, &in_base_uri, &abs_uri);
-			bool rooted = uri_is_under(&writer->base_uri, &writer->root_uri);
-			SerdURI* root = rooted ? &writer->root_uri : & writer->base_uri;
-			if (!uri_is_under(&abs_uri, root) ||
-			    writer->syntax == SERD_NTRIPLES ||
-			    writer->syntax == SERD_NQUADS) {
-				serd_uri_serialise(&abs_uri, uri_sink, writer);
-			} else {
-				serd_uri_serialise_relative(
-					&uri, &writer->base_uri, root, uri_sink, writer);
-			}
-		} else {
-			write_uri(writer, node->buf, node->n_bytes);
-		}
-		sink(">", 1, writer);
-		if (is_inline_start(writer, field, flags)) {
-			sink(" ;", 2, writer);
-			write_newline(writer);
-		}
-	default:
+		ret = write_uri_node(writer, node, field, flags);
 		break;
+	case SERD_CURIE:
+		ret = write_curie(writer, node, field, flags);
+		break;
+	case SERD_BLANK:
+		ret = write_blank(writer, node, field, flags);
+	default: break;
 	}
 	writer->last_sep = SEP_NONE;
-	return true;
+	return ret;
 }
 
 static inline bool
@@ -673,7 +707,7 @@ serd_writer_write_statement(SerdWriter*        writer,
 	}
 
 #define TRY(write_result) \
-	if (!write_result) { \
+	if (!(write_result)) { \
 		return SERD_ERR_UNKNOWN; \
 	}
 
@@ -823,14 +857,12 @@ SerdStatus
 serd_writer_finish(SerdWriter* writer)
 {
 	if (writer->context.subject.type) {
-		sink(" .\n", 3, writer);
+		write_sep(writer, SEP_END_S);
 	}
 	if (writer->context.graph.type) {
-		sink("}\n", 2, writer);
-	}
-	if (writer->style & SERD_STYLE_BULK) {
-		serd_bulk_sink_flush(&writer->bulk_sink);
+		write_sep(writer, SEP_GRAPH_END);
 	}
+	serd_byte_sink_flush(&writer->byte_sink);
 	writer->indent = 0;
 	return free_context(writer);
 }
@@ -853,14 +885,11 @@ serd_writer_new(SerdSyntax     syntax,
 	writer->root_uri     = SERD_URI_NULL;
 	writer->base_uri     = base_uri ? *base_uri : SERD_URI_NULL;
 	writer->anon_stack   = serd_stack_new(sizeof(WriteContext));
-	writer->sink         = ssink;
-	writer->stream       = stream;
 	writer->context      = context;
 	writer->list_subj    = SERD_NODE_NULL;
 	writer->empty        = true;
-	if (style & SERD_STYLE_BULK) {
-		writer->bulk_sink = serd_bulk_sink_new(ssink, stream, SERD_PAGE_SIZE);
-	}
+	writer->byte_sink    = serd_byte_sink_new(
+		ssink, stream, (style & SERD_STYLE_BULK) ? SERD_PAGE_SIZE : 1);
 	return writer;
 }
 
@@ -959,9 +988,7 @@ serd_writer_free(SerdWriter* writer)
 	serd_writer_finish(writer);
 	serd_stack_free(&writer->anon_stack);
 	free(writer->bprefix);
-	if (writer->style & SERD_STYLE_BULK) {
-		serd_bulk_sink_free(&writer->bulk_sink);
-	}
+	serd_byte_sink_free(&writer->byte_sink);
 	serd_node_free(&writer->root_node);
 	free(writer);
 }
