@@ -1,5 +1,5 @@
 /*
-  Copyright 2011-2016 David Robillard <http://drobilla.net>
+  Copyright 2011-2017 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -87,6 +87,40 @@ test_sink(void*              handle,
 	return SERD_SUCCESS;
 }
 
+static int
+check_file_uri(const char* hostname,
+               const char* path,
+               bool        escape,
+               const char* expected_uri,
+               const char* expected_path)
+{
+	if (!expected_path) {
+		expected_path = path;
+	}
+
+	SerdNode node = serd_node_new_file_uri(
+		USTR(path), USTR(hostname), 0, escape);
+
+	uint8_t* out_hostname = NULL;
+	uint8_t* out_path     = serd_file_uri_parse(node.buf, &out_hostname);
+	int      ret          = 0;
+	if (strcmp((const char*)node.buf, expected_uri)) {
+		ret = failure("Bad URI %s != %s\n", node.buf, expected_uri);
+	} else if (!hostname && out_hostname) {
+		ret = failure("hostname `%s' shouldn't exist\n", out_hostname);
+	} else if (hostname && !out_hostname) {
+		ret = failure("expected hostname `%s'\n", hostname);
+	} else if (strcmp((const char*)out_path, (const char*)expected_path)) {
+		ret = failure("path=>URI=>path failure %s => %s => %s\n",
+		              path, node.buf, out_path);
+	}
+
+	free(out_path);
+	free(out_hostname);
+	serd_node_free(&node);
+	return ret;
+}
+
 int
 main(void)
 {
@@ -269,111 +303,27 @@ main(void)
 		return failure("Bad path %s for %s\n", serd_uri_to_path(uri), uri);
 	}
 
-	// Test serd_node_new_file_uri and serd_file_uri_parse
-	SerdURI        furi;
-	const uint8_t* path_str  = USTR("C:/My 100%");
-	SerdNode       file_node = serd_node_new_file_uri(path_str, 0, &furi, true);
-	uint8_t*       hostname  = NULL;
-	uint8_t*       out_path  = serd_file_uri_parse(file_node.buf, &hostname);
-	if (strcmp((const char*)file_node.buf, "file:///C:/My%20100%%")) {
-		return failure("Bad URI %s\n", file_node.buf);
-	} else if (hostname) {
-		return failure("hostname `%s' shouldn't exist\n", hostname);
-	} else if (strcmp((const char*)path_str, (const char*)out_path)) {
-		return failure("path=>URI=>path failure %s => %s => %s\n",
-		               path_str, file_node.buf, out_path);
-	}
-	free(out_path);
-	serd_node_free(&file_node);
+	// Test file URI escaping and parsing
 
-	path_str  = USTR("C:\\Pointless Space");
-	file_node = serd_node_new_file_uri(path_str, USTR("pwned"), 0, true);
-	hostname  = NULL;
-	out_path  = serd_file_uri_parse(file_node.buf, &hostname);
-	if (strcmp((const char*)file_node.buf, "file://pwned/C:/Pointless%20Space")) {
-		return failure("Bad URI %s\n", file_node.buf);
-	} else if (!hostname || strcmp((const char*)hostname, "pwned")) {
-		return failure("Bad hostname `%s'\n", hostname);
-	} else if (strcmp((const char*)out_path, "C:/Pointless Space")) {
-		return failure("path=>URI=>path failure %s => %s => %s\n",
-		               path_str, file_node.buf, out_path);
-	}
-	free(hostname);
-	free(out_path);
-	serd_node_free(&file_node);
-
-	path_str  = USTR("/foo/bar");
-	file_node = serd_node_new_file_uri(path_str, 0, 0, true);
-	hostname  = NULL;
-	out_path  = serd_file_uri_parse(file_node.buf, &hostname);
-	if (strcmp((const char*)file_node.buf, "file:///foo/bar")) {
-		return failure("Bad URI %s\n", file_node.buf);
-	} else if (hostname) {
-		return failure("hostname `%s' shouldn't exist\n", hostname);
-	} else if (strcmp((const char*)path_str, (const char*)out_path)) {
-		return failure("path=>URI=>path failure %s => %s => %s\n",
-		               path_str, file_node.buf, out_path);
-	}
-	free(out_path);
-	serd_node_free(&file_node);
-
-	path_str  = USTR("/foo/bar");
-	file_node = serd_node_new_file_uri(path_str, USTR("localhost"), 0, true);
-	out_path  = serd_file_uri_parse(file_node.buf, &hostname);
-	if (strcmp((const char*)file_node.buf, "file://localhost/foo/bar")) {
-		return failure("Bad URI %s\n", file_node.buf);
-	} else if (strcmp((const char*)hostname, "localhost")) {
-		return failure("incorrect hostname `%s'\n", hostname);
-	} else if (strcmp((const char*)path_str, (const char*)out_path)) {
-		return failure("path=>URI=>path failure %s => %s => %s\n",
-		               path_str, file_node.buf, out_path);
-	}
-	free(hostname);
-	free(out_path);
-	serd_node_free(&file_node);
-
-	// File URI with space and no escaping
-	path_str  = USTR("a/relative path");
-	file_node = serd_node_new_file_uri(path_str, 0, 0, false);
-	out_path  = serd_file_uri_parse(file_node.buf, &hostname);
-	if (strcmp((const char*)file_node.buf, "a/relative path")) {
-		return failure("Bad URI %s\n", file_node.buf);
-	} else if (hostname) {
-		return failure("hostname `%s' shouldn't exist\n", hostname);
-	} else if (strcmp((const char*)path_str, (const char*)out_path)) {
-		return failure("path=>URI=>path failure %s => %s => %s\n",
-		               path_str, file_node.buf, out_path);
-	}
-	free(hostname);
-	free(out_path);
-	serd_node_free(&file_node);
-
-	// File URI with space and escaping
-	path_str  = USTR("a/relative path");
-	file_node = serd_node_new_file_uri(path_str, 0, 0, false);
-	out_path  = serd_file_uri_parse(file_node.buf, &hostname);
-	if (strcmp((const char*)file_node.buf, "a/relative path")) {
-		return failure("Bad URI %s\n", file_node.buf);
-	} else if (hostname) {
-		return failure("hostname `%s' shouldn't exist\n", hostname);
-	} else if (strcmp((const char*)path_str, (const char*)out_path)) {
-		return failure("path=>URI=>path failure %s => %s => %s\n",
-		               path_str, file_node.buf, out_path);
-	}
-	free(hostname);
-	free(out_path);
-	serd_node_free(&file_node);
-
-	if (serd_file_uri_parse(USTR("file://invalid"), NULL)) {
-		return failure("successfully parsed bogus URI <file://invalid>\n");
+	if (check_file_uri(NULL, "C:/My 100%", true,
+	                   "file:///C:/My%20100%%", NULL) ||
+	    check_file_uri("ahost", "C:\\Pointless Space", true,
+	                   "file://ahost/C:/Pointless%20Space",
+	                   "C:/Pointless Space") ||
+	    check_file_uri(NULL, "/foo/bar", true,
+	                   "file:///foo/bar", NULL) ||
+	    check_file_uri("bhost", "/foo/bar", true,
+	                   "file://bhost/foo/bar", NULL) ||
+	    check_file_uri(NULL, "a/relative path", false,
+	                   "a/relative path", NULL) ||
+	    check_file_uri(NULL, "a/relative <path>", true,
+	                   "a/relative%20%3Cpath%3E", NULL)) {
+		return 1;
 	}
 
-	out_path = serd_file_uri_parse(USTR("file://host/foo/%XYbar"), NULL);
-	if (strcmp((const char*)out_path, "/foo/bar")) {
-		return failure("bad tolerance of junk escape: `%s'\n", out_path);
-	}
-	free(out_path);
-	out_path = serd_file_uri_parse(USTR("file://host/foo/%0Abar"), NULL);
+	// Test tolerance of parsing junk URI escapes
+
+	uint8_t* out_path = serd_file_uri_parse(USTR("file:///foo/%0Xbar"), NULL);
 	if (strcmp((const char*)out_path, "/foo/bar")) {
 		return failure("bad tolerance of junk escape: `%s'\n", out_path);
 	}
@@ -416,6 +366,27 @@ main(void)
 		return failure("Creating node from NULL string failed\n");
 	}
 
+	// Test serd_node_from_substring
+
+	SerdNode empty = serd_node_from_substring(SERD_LITERAL, NULL, 32);
+	if (empty.buf || empty.n_bytes || empty.n_chars || empty.flags || empty.type) {
+		return failure("Successfully created node from NULL substring\n");
+	}
+
+	SerdNode a_b = serd_node_from_substring(SERD_LITERAL, USTR("a\"bc"), 3);
+	if (a_b.n_bytes != 3 || a_b.n_chars != 3 || a_b.flags != SERD_HAS_QUOTE
+	    || strncmp((const char*)a_b.buf, "a\"b", 3)) {
+		return failure("Bad node %s %zu %zu %d %d\n",
+		               a_b.buf, a_b.n_bytes, a_b.n_chars, a_b.flags, a_b.type);
+	}
+
+	a_b = serd_node_from_substring(SERD_LITERAL, USTR("a\"bc"), 10);
+	if (a_b.n_bytes != 4 || a_b.n_chars != 4 || a_b.flags != SERD_HAS_QUOTE
+	    || strncmp((const char*)a_b.buf, "a\"bc", 4)) {
+		return failure("Bad node %s %zu %zu %d %d\n",
+		               a_b.buf, a_b.n_bytes, a_b.n_chars, a_b.flags, a_b.type);
+	}
+
 	// Test serd_node_new_uri_from_string
 
 	SerdNode nonsense = serd_node_new_uri_from_string(NULL, NULL, NULL);
@@ -538,17 +509,11 @@ main(void)
 
 	if (!serd_writer_set_base_uri(writer, &lit)) {
 		return failure("Set base URI to %s\n", lit.buf);
-	}
-
-	if (!serd_writer_set_prefix(writer, &lit, &lit)) {
+	} else if (!serd_writer_set_prefix(writer, &lit, &lit)) {
 		return failure("Set prefix %s to %s\n", lit.buf, lit.buf);
-	}
-
-	if (!serd_writer_end_anon(writer, NULL)) {
+	} else if (!serd_writer_end_anon(writer, NULL)) {
 		return failure("Ended non-existent anonymous node\n");
-	}
-
-	if (serd_writer_get_env(writer) != env) {
+	} else if (serd_writer_get_env(writer) != env) {
 		return failure("Writer has incorrect env\n");
 	}
 
@@ -566,6 +531,7 @@ main(void)
 	                              { &SERD_NODE_NULL, &p, &o, NULL, NULL },
 	                              { &s, &o, &o, NULL, NULL },
 	                              { &o, &p, &o, NULL, NULL },
+	                              { &s, &p, &SERD_NODE_NULL, NULL, NULL },
 	                              { NULL, NULL, NULL, NULL, NULL } };
 	for (unsigned i = 0; i < sizeof(junk) / (sizeof(SerdNode*) * 5); ++i) {
 		if (!serd_writer_write_statement(
@@ -595,10 +561,16 @@ main(void)
 		}
 	}
 
-	// Write 1 statement with bad UTF-8 (should be replaced)
+	// Write statements with bad UTF-8 (should be replaced)
+	const uint8_t bad_str[] = { 0xFF, 0x90, 'h', 'i', 0 };
+	SerdNode      bad_lit   = serd_node_from_string(SERD_LITERAL, bad_str);
+	SerdNode      bad_uri   = serd_node_from_string(SERD_URI, bad_str);
 	if (serd_writer_write_statement(writer, 0, NULL,
-	                                &s, &p, &o, NULL, NULL)) {
-		return failure("Failed to write junk UTF-8\n");
+	                                &s, &p, &bad_lit, NULL, NULL)) {
+		return failure("Failed to write junk UTF-8 literal\n");
+	} else if (serd_writer_write_statement(writer, 0, NULL,
+	                                       &s, &p, &bad_uri, NULL, NULL)) {
+		return failure("Failed to write junk UTF-8 URI\n");
 	}
 
 	// Write 1 valid statement
@@ -650,19 +622,16 @@ main(void)
 
 	if (!serd_reader_read_file(reader, USTR("http://notafile"))) {
 		return failure("Apparently read an http URI\n");
-	}
-	if (!serd_reader_read_file(reader, USTR("file:///better/not/exist"))) {
+	} else if (!serd_reader_read_file(reader, USTR("file:///better/not/exist"))) {
 		return failure("Apparently read a non-existent file\n");
-	}
-	if (!serd_reader_read_file(reader, USTR("file://"))) {
+	} else if (!serd_reader_read_file(reader, USTR("file://"))) {
 		return failure("Apparently read a file with no path\n");
 	}
-	SerdStatus st = serd_reader_read_file(reader, USTR(path));
+
+	const SerdStatus st = serd_reader_read_file(reader, USTR(path));
 	if (st) {
 		return failure("Error reading file (%s)\n", serd_strerror(st));
-	}
-
-	if (rt->n_statements != 12) {
+	} else if (rt->n_statements != 13) {
 		return failure("Bad statement count %d\n", rt->n_statements);
 	} else if (!rt->graph || !rt->graph->buf ||
 	           strcmp((const char*)rt->graph->buf, "http://example.org/")) {
